/* $Header: /HBFontEdit/FontChange.c 2     15-11-96 18:25 Bert $ */
/* $Nokeywords: $ */

#include <windows.h>

#include "fontchar.h"
#include "fontedit.h"
#include "fonthead.h"
#include "fontchange.h"  // #include "fonttypes.h"
#include "FontCvt.h"

#include "FontVars.h"

//**************************************************************************
//              Local Variables                                             
//**************************************************************************
HDC hNewMemDC;
HDC hBoxMemDC;
HBITMAP hNewBitmap;
HBITMAP hBoxBitmap;
HCURSOR hOldCursor = NULL;      // Holds Arrow while we show Hourglass 
HCURSOR hHourGlass = NULL;
DWORD oldMode;                   // For StretchBltMode 
WORD cHeader, cTable;

#define ALPHA_CNT        26
#define TOTAL_WEIGHTS  1000
#define FSF_FIXED             0x0001
#define FSF_PROPORTIONAL      0x0002
#define WORD_LIMIT    65534    /* max. number of bytes that can be read(written) by */


//***************************************************************************
// * BoxToClipboard(ptA, width, height)
// *
// * purpose: write char (or part of it) to clipboard
// *
// * params : POINT ptA    : upper left coordinate
// *          DWORD width  : width of char in pixels
// *          DWORD height : height of char in pixels
// * returns: none
// *
// * side effects: none
// *
// ***************************************************************************
void
BoxToClipboard(FontType* lpFnt,
    POINT ptA,                               // Upper left point 
    DWORD width,
    DWORD height                      // Size 
    )
{
    HDC hDC;
    DWORD x, y;

    hDC = GetDC(hFont);                         // DC to be compatible with 
    hNewBitmap = CreateBitmap(
                    width, height,
                    1, 1, (LPBYTE)NULL);
    hNewMemDC = CreateCompatibleDC(hDC);                // Create a DC 
    SelectObject(hNewMemDC, hNewBitmap);                // Relate them 
    ReleaseDC(hFont, hDC);                      // Done with font DC 

    for (x = 0; x < width; x++)
        for (y = 0; y < height; y++)
            SetPixel(hNewMemDC, x, y, matBox[x + ptA.x][y + ptA.y] == TRUE ?
            BLACK : WHITE);

    // Now wake up Clipboard and empty it 
    if (!OpenClipboard(hFont))
        ErrorBox(hBox, vszCannotOpenClip); // , vszCopyingToClip);
    else        // Ok: We got the Clipboard 
        {
        EmptyClipboard();
        SetClipboardData(CF_BITMAP, hNewBitmap);        // Tell Clipboard 
        }

    // Tidy things up 
    CloseClipboard();
    DeleteDC(hNewMemDC);
}


//***************************************************************************
// * WORD ClipboardToBox(ptA, width, height, fFit)
// *
// * purpose: copies char (or part of char ) from clipboard to work bitmap
// *          stretching it if need be
// *
// * params : PIONT ptA    : upper left coordinate
// *          DWORD width  : width of char in pixels
// *          DWORD height : height of char in pixels
// *          BOOL fFit   : flag to indicate if default width is to be used
// * returns: none
// *
// * side effects: pixel values of bitmap may change for char
// *
// ***************************************************************************

DWORD ClipboardToBox(FontType* lpFnt,
    POINT ptA,                               // Upper left point 
    DWORD width,
    DWORD height,                     // Size 
    BOOL fFit                               // Use default width if TRUE 
    )
{
    BITMAP bitmap;
    HDC hDC;
    DWORD x, y;
    HANDLE hT;

    if (!OpenClipboard(hFont)) {
        ErrorBox(hBox, vszCannotOpenClip);
        return 0;
    }
    hNewBitmap = GetClipboardData(CF_BITMAP);

    // Check if we got something like a character 
    if (GetObject(hNewBitmap, sizeof(BITMAP), (LPSTR)&bitmap) != sizeof(BITMAP))
        {   // What did we get 
        ErrorBox(hBox, vszErrorClip);
        CloseClipboard();
        return 0;
    }


    if (fFit && ((WORD)bitmap.bmWidth <= lpFnt->Header.MaxWidth))
        width = bitmap.bmWidth;

    hDC = GetDC(hFont);                         // DC to be compatible with 
    hBoxBitmap = CreateBitmap(
                    width, height,
                    1, 1, (LPBYTE)NULL);
    hBoxMemDC = CreateCompatibleDC(hDC);                // Create a DC 
    hT = SelectObject(hBoxMemDC, hBoxBitmap);           // Relate them 
    if (hT == NULL || hDC == NULL || hBoxBitmap == NULL || hBoxMemDC == NULL) {
        ErrorBox(hBox, vszErrorClip);
	CloseClipboard();
	DeleteDC(hBoxMemDC);
	DeleteObject(hBoxBitmap);
        return 0;
    }

    // Get a DC to relate to the Bitmap we just got 
    hNewMemDC = CreateCompatibleDC(hDC);                // Create a DC 
    hT = SelectObject(hNewMemDC, hNewBitmap);           // Relate them 
    ReleaseDC(hFont, hDC);                      // Done with font DC 
    if (hT == NULL || hNewMemDC == NULL) {
        ErrorBox(hBox, vszErrorClip);
	CloseClipboard();
	DeleteDC(hNewMemDC);
	DeleteDC(hBoxMemDC);
	DeleteObject(hBoxBitmap);
        return 0;
    }

    // Now StretchBlt whatever was on the clipboard into the character 
    oldMode = SetStretchBltMode(hBoxMemDC, COLORONCOLOR);
    fFit = StretchBlt(hBoxMemDC, 0, 0, width, height,
                hNewMemDC, 0, 0, bitmap.bmWidth, bitmap.bmHeight, SRCCOPY);
    if (!fFit || !oldMode) {
        ErrorBox(hBox, vszErrorClip);
	CloseClipboard();
	DeleteDC(hNewMemDC);
	DeleteDC(hBoxMemDC);
	DeleteObject(hBoxBitmap);
        return 0;
    }
    (void)SetStretchBltMode(hBoxMemDC, oldMode);
    for (x = 0; x < width; x++)
        for (y = 0; y < height; y++)
            matBox[x + ptA.x] [y + ptA.y] = (CHAR)(GetPixel(hBoxMemDC, x, y) ?
                0 : 1);
    // Tidy things up 
    DeleteDC(hNewMemDC);
    DeleteDC(hBoxMemDC);
    DeleteObject(hBoxBitmap);
    CloseClipboard();
    return(width);
}


//***************************************************************************
// * ToClipboard(iChar, width, height)
// *
// * purpose: write char in edit box to clipboard
// *
// * params : BYTE iChar  : number of char to be copied to clipboard
// *          DWORD width  : width of char in pixels
// *          DWORD height : height of char in pixels
// *
// * returns: none
// *
// * side effects: none
// *
// ***************************************************************************

void
ToClipboard(FontType *lpFnt,
    BYTE iChar,
    DWORD width,             // Here in Pixels 
    DWORD height             // Also in Pixels 
    )
{
    HDC hDC;

    hDC = GetDC(hFont);                         // DC to be compatible with 
    hNewBitmap = CreateBitmap(
                    width,                      // Width of font in pixels 
                    height,
                    1, 1, (LPBYTE)NULL);
    hNewMemDC = CreateCompatibleDC(hDC);                // Create a DC 
    SelectObject(hNewMemDC, hNewBitmap);                // Relate them 
    ReleaseDC(hFont, hDC);                      // Done with font DC 

    BitBlt(hNewMemDC, 0, 0, width, height,      // Move Character in 
            lpFnt->BmDC, lpFnt->offsets[iChar], 0, NOTSRCCOPY);

    // Now wake up Clipboard and empty it 
    if (!OpenClipboard(hFont))
        ErrorBox(hBox, vszCannotOpenClip); // , vszCopyingToClip);
    else        // Ok: We got the Clipboard 
        {
        EmptyClipboard();
        SetClipboardData(CF_BITMAP, hNewBitmap);        // Tell Clipboard 
        }

    // Tidy things up 
    CloseClipboard();
    DeleteDC(hNewMemDC);
}


