/* $Header: /HBFontEdit/fontload.c 7     15-11-96 18:25 Bert $ */
/* $Nokeywords: $ */

#include <windows.h>

//#include "fontchar.h"
#include "fontedit.h"
//#include "fonthead.h"
#include "fontload.h"  // #include "fonttypes.h"
#include "FontCvt.h"


#define BLACK 0x000000
#define WHITE 0xFFFFFF
//#include "FontVars.h"
//**************************************************************************
//              Local Variables                                             
//**************************************************************************
typedef struct
{
	char*	lpFontBody;
	char*	lpWork;

	BYTE*	lp1;	// Pointers to bitmaps in format N 
	BYTE*	lp2;

	HBITMAP hNewBitmap;
	HBITMAP hBoxBitmap;
	//HCURSOR hOldCursor;// = NULL;      // Holds Arrow while we show Hourglass 
	//HCURSOR hHourGlass;// = NULL;

	FontHeaderType fontBuffer;
} WorkBufferType;


typedef struct
{
	WORD cHeader, cTable;
	DWORD oldMode;                   // For StretchBltMode 

	HDC hNewMemDC;
	HDC hBoxMemDC;
	HBITMAP hNewBitmap;
} ResizeBufferType;

//FontHeaderType lpWrk->fontBuffer;

DWORD Proport(DWORD, DWORD, DWORD, DWORD);  // Reproportions a value 

//CHAR  *lpFontBody = NULL;  // Pointer to Font Body 
//CHAR  *lpWork = NULL;      // Pointer to Work Area 

#define ALPHA_CNT        26
#define TOTAL_WEIGHTS  1000
#define FSF_FIXED             0x0001
#define FSF_PROPORTIONAL      0x0002
#define WORD_LIMIT    65534    /* max. number of bytes that can be read(written) by */

//**************************************************************************
//              Local Functions                                             
//**************************************************************************

void NewAverage(FontType* lpFnt);
BOOL GetNewMap(FontType* lpFnt, ResizeBufferType *lpRsz, DWORD width, DWORD height,HWND equalHWND);
void UseNewMap(FontType* lpFnt, ResizeBufferType *lpRsz);
void ShrinkFont(FontType *lpFnt, char first, char last, HWND equalHWND);
// ****************************************************************************
// *WORD ConvertToBitmapFormat(width, phase, height)
// *
// * purpose  : Takes a part of the font file and converts it to a string of
// *            bytes for a single scan of bitmap for a character.
// *
// * params   : WORD width : width of the character in pixels(bits)
// *            WORD phase : current deviation from byte alignment (pixels)
// *            WORD height: height of character(pixels)
// *
// * returns  : WORD phase : new deviation from byte alignment (pixels)
// *
// * side effects : modifies pointers lp1 and lp2 (pointing to work space and
// *                font body respectively)
// *
// ***************************************************************************
DWORD PASCAL
ConvertToBitmapFormat(WorkBufferType *lpWrk,
     DWORD width,     // width of the character in pixels(bits) 
     DWORD phase,     // current deviation from byte alignment (pixels) 
     DWORD height     // height of character(pixels) 
     )
{
      INT w;
      WORD j;

      // in the font file format characters are stored consecutively in
      //   column-major ordering (columns of char 1 followed by columns of char
      //   2 ... etc.). lp2 points to start of columns of current character.
      //   lp1 points to start of row of bitmap for character. 

      for (w = width; w > 0; w -= 8){
          if (phase == 0){
          //  easy case 
              *lpWrk->lp1++ = *lpWrk->lp2;

              if (w < 8)
                  phase = w;
          }
          else{

              --lpWrk->lp1;
              j = (WORD)*lpWrk->lp1;
              j <<= 8;
              j |= (((WORD)*lpWrk->lp2) << (8 - phase));
              *lpWrk->lp1++ = (BYTE)(j >> 8);
              *lpWrk->lp1++ = (BYTE)j;
              if (w < 8){
                  phase += w;
                  if (phase <= 8)
                      lpWrk->lp1--;          // back up pointer 
                  phase &= 7;
              }
          }
          lpWrk->lp2 += height;          // move to next column 
      }
      return phase;
}

//***************************************************************************
// * char * VerifyTableContents()
// *
// * purpose  : scan the offsets table of file just read and check if
// *            width and offsets lie within limits
// *
// * params    : none
// *
// * returns   : char * szError : ptr to Error message if table not OK
// *                              NULL otherwise
// *
// * side effects : none
// *
// ***************************************************************************
CHAR * PASCAL VerifyTableContents(FontType *lpFnt,WorkBufferType *lpWrk)
{
	 PBYTE		  pjGlyphData;
     GLYPHINFO_20 gi2T20;        // temp ptr to a 2.0 style table
     GLYPHINFO_30 gi3T30;        // temp ptr to a 2.0 style table
     INT i;

     // separate loops written for the 2.0 and 3.0 font processing because
     //   a single loop would involve too many checks (of font type) within
     //   the loop and slow down processing  
     if (lpFnt->FontFormat == ID_FORMAT2){
         pjGlyphData = (lpWrk->lpFontBody +1);

		// view table as 2.0 table 

         // check that each and every width and offset lie within limits 

         for (i=0; i < (lpWrk->fontBuffer.LastChar - lpWrk->fontBuffer.FirstChar + 1); i++) {

			  vGlyphInfo20FromBuffer (pjGlyphData, &gi2T20);

              if (gi2T20.GIwidth > 64)
                  return vszTableWidthsBad;

              if ((gi2T20.GIoffset > (UINT)WORD_LIMIT) ||
                  (gi2T20.GIoffset < 0) )
                  return vszTableOffsetsBad;

			  pjGlyphData += lSizeOfOldGlyph20;
         }
     }
     else{

         pjGlyphData = lpWrk->lpFontBody;

            // view table as 3.0 table 
         // check that each and every width and offset lie within limits 

         for (i=0; i< (lpWrk->fontBuffer.LastChar - lpWrk->fontBuffer.FirstChar +1); i++) {

			  vGlyphInfo30FromBuffer (pjGlyphData, &gi3T30);

              if (gi3T30.GIwidth > 64)
                  return vszTableWidthsBad;

              if (gi3T30.GIoffset < (DWORD)0)
                  return vszTableOffsetsBad;

			  pjGlyphData += lSizeOfOldGlyph30;
         }
     }
     return NULL;
}

//***************************************************************************
// * char * FontLoad()
// *
// * purpose: reads in the specified font file and creates a work bitmap for
// *          the editor. Also creates a local bitmap-offset table for easy
// *          access to the characters in the bitmap
// *
// * params : pszFileName - File Name to open.
// *
// * returns: ptr to NULL string if Open goes off OK
// *          ptr to error message string otherwise
// *
// * side effects: lots
// *
// ***************************************************************************
CHAR * FontLoad(FontType *lpFnt, char *pszFileName, HWND equalHWND)
{
	WorkBufferType work;
    CHAR  *lpFontBodySav = NULL;  // local pointer to font body 
    DWORD len, fontlen;               // length of font body (bytes)
	UINT i;
  
    DWORD iWorkLen;                   // length of work area (bytes) 
    HDC hDC;
    CHAR * pszError;                   // error msg if header is messed up 

    DWORD row, height, width,phase;
    DWORD offset;
    GLYPHINFO_20 gi2GlyphTable20;    // Pointer into 2.0 style offsets table 
    GLYPHINFO_30 gi3GlyphTable30;    // Pointer into 3.0 style offsets table 
	PBYTE 	pjGlyphData;			 // Pointer to Glyph information buffer. 

    BYTE  cDummy [CCHEXTRA];         // dummy buffer for unneeded 3.0 header
                                     //   info
                                    
	HANDLE nFile;
	DWORD BytesRead;

	// Put up an hourglass ... this may take a while 
    if (!hHourGlass)
        hHourGlass = LoadCursor(NULL, IDC_WAIT);        // Get Hourglass 
    hOldCursor = SetCursor(hHourGlass);         // Show hourglass 

    // open file for read. 
	nFile = CreateFile(pszFileName, GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	
	if (nFile == INVALID_HANDLE_VALUE) 
	{

		return vszErrorOpeningFile;
	}

	//- ReadFile:  Here is where we need to make some adjustments in order to
	//- read the file in to a new DWORD aligned format.
	{
		BYTE    jBuffer [200];


		// Read the header 
		if ((!ReadFile(nFile, (LPSTR)&jBuffer, lSizeOfOldFontHeader,&BytesRead,NULL))
			 || ((BytesRead) != (DWORD)lSizeOfOldFontHeader) )
		{
			CloseHandle(nFile);
			return vszErrorReadingHdr;
		}

		//
		// Call conversion routine to give us a properly aligned buffer.
		//

		vFontStructFromBuffer (
				 (PBYTE)&jBuffer,
				 &work.fontBuffer
				);

	}

    // Check the version number -- make sure it's a font 
    switch (work.fontBuffer.Version)
        {
        case 0x200:
				work.fontBuffer.Version = ID_FORMAT2;
                break;
        case 0x300:        // added 1/19/88 
                work.fontBuffer.Version = ID_FORMAT3;
                // read in the extra fields into a dummy buffer ... they don't
                // * mean anything to us  at this stage.
                 
                if ((!ReadFile(nFile, (LPSTR)cDummy, CCHEXTRA,&BytesRead,NULL))
					|| (BytesRead != CCHEXTRA))
				{
					CloseHandle(nFile);
                    return vszErrorReadingHdr;
                }

                break;
        default:                // Anything else -- toughies 
				CloseHandle(nFile);
                return vszUnknownFormat;
        }

    // check if contents of font header are sensible. If not,
    //   give an error message and quit 

    if ((pszError = VerifyHeaderContents(&lpFnt->Header)) != NULL) 
	{
		CloseHandle(nFile);
        return pszError;
	}

    // Ready to load -- Check if we will be overwriting 
    if (lpFnt->FontFormat)
	{
        DeleteGlobalBitmap(lpFnt);
        //fLoaded = FALSE;
	}

    // Allocate space for font body and Face Name and read them in 
    len= (work.fontBuffer.Size - lSizeOfOldFontHeader);  // Compute size 

    if ((work.lpFontBody = (LPSTR)GlobalAlloc(GMEM_ZEROINIT, (LONG)len)) == 0) 
	{
		CloseHandle(nFile);
        return vszNotEnoughMem;                       // Get Handle to space 
	}

    if (work.fontBuffer.Version == ID_FORMAT3)
        fontlen = len - CCHEXTRA;
    else
        fontlen = len;


    // read the partially filled segment  
    if ((!ReadFile(nFile, work.lpFontBody, (DWORD)fontlen,&BytesRead,NULL)) ||
		(BytesRead != fontlen)) 
	{
        GlobalFree (work.lpFontBody);
		CloseHandle (nFile);
        return vszErrorReadingBody;
	}

    // Close the file 
	CloseHandle (nFile);

    // check if the offset table entries are within allowable limits.
    //   If not give an error message, clean up and quit 
    if ((pszError = VerifyTableContents(lpFnt,&work)) != NULL) {

        GlobalFree (work.lpFontBody);
        return pszError;
	}

    // now that everything has been checked, move buffer to font 
    lpFnt->Header = work.fontBuffer;
	lpFnt->CharCount = 1+ lpFnt->Header.LastChar - lpFnt->Header.FirstChar;
	lpFnt->FontFormat = work.fontBuffer.Version;
    // Make local copies of FaceName and Device Name 
    if (lpFnt->Header.Face){

        lstrcpy((LPSTR)lpFnt->szFaceName, work.lpFontBody + lpFnt->Header.Face -
                                         (lpFnt->FontFormat == ID_FORMAT2 ?
                                                    lSizeOfOldFontHeader   :
                                                    lSizeOfOldFontHeader30-1));
    } else {

        lstrcpy((LPSTR)lpFnt->szFaceName, (LPSTR)"");
	}

    for (i = 0; i <= 256; i++)      // Zero offsets Table for below 
        lpFnt->offsets[i] = 0;

    // compute work space needed if a 3.0 file. This has to be done since
    //   3.0 files are compressed versions of 2.0 files and may need a
    //   work bitmap bigger than the actual font body size 

    if (lpFnt->FontFormat == ID_FORMAT3) 
	{
        cTable = (WORD) (6 * (lpFnt->CharCount + 1));
        iWorkLen = 0;

        pjGlyphData = work.lpFontBody;

        // work bitmap size = sum of sizes of all characters in font 

        for (i = lpFnt->Header.FirstChar; i <= lpFnt->Header.LastChar; i++) {

			 vGlyphInfo30FromBuffer (pjGlyphData, &gi3GlyphTable30);

             iWorkLen += ((gi3GlyphTable30.GIwidth +7) >> 3) * lpFnt->Header.PixHeight;

			 pjGlyphData += lSizeOfOldGlyph30;
		}

    } else { // 2.0 file 

        cTable = (WORD)(4 * (lpFnt->CharCount + 1));
           // compute table length 
        iWorkLen = len;  // work space for a 2.0 file is the same as the
                         //    length of the font body 
    }

	//- Add some extra space for dword alignment.
	iWorkLen += (lpFnt->Header.PixHeight * sizeof (DWORD));
      // Get work space 

    if ((work.lpWork = (LPSTR)GlobalAlloc (GMEM_ZEROINIT, (LONG)iWorkLen)) == 0) {

        GlobalFree (work.lpFontBody);
        return vszNotEnoughMem;
    }

	work.lp1 = work.lpWork;

    height = (DWORD) lpFnt->Header.PixHeight;
    offset = 0;
    // put the font file into bitmap format 
    if (lpFnt->FontFormat == ID_FORMAT2){         // table in 2.0 format 
        for (row = 0; row < height; row++){

             // view table as a 2.0 style table 
             pjGlyphData = work.lpFontBody + 1;

             phase = 0;

             for (i = 0; i < (UINT)(lpFnt->CharCount); i++) {

				vGlyphInfo20FromBuffer (pjGlyphData, &gi2GlyphTable20);

                width = (DWORD) gi2GlyphTable20.GIwidth;

                // size of each table element = 4bytes 
                work.lp2 = work.lpFontBody + (gi2GlyphTable20.GIoffset -
							lSizeOfOldFontHeader) + row;

				pjGlyphData += lSizeOfOldGlyph20;

                // offset ends up as the sum of the widths 
                if (row == 0)              // Once is enough 
						lpFnt->offsets[i + lpFnt->Header.FirstChar + 1] = offset += width;

                // create a single scan of bitmap for character 
                phase = ConvertToBitmapFormat (&work,width, phase, height);
             }
             if ((work.lp1 - work.lpWork) & 1)
                 *work.lp1++ = 0;             // Round lp1 up to Word Boundary 
#ifdef DWORDROUND
             if ((work.lp1 - work.lpWork) & 2) {
                 *work.lp1++ = 0;             // Round lp1 up to DWord Boundary 
                 *work.lp1++ = 0;             // Round lp1 up to DWord Boundary 
			 }
#endif
             //if (((offset + 7) >> 3) & 1)
                 ///lp1++ = 0;             // Round lp1 up to Word Boundary 
             //if (((offset + 7) >> 3) & 2) {
                 ///lp1++ = 0;             // Round lp1 up to DWord Boundary 
                 ///lp1++ = 0;             // Round lp1 up to DWord Boundary 
			 //}
        }
    }
     // separate loops written for the 2.0 and 3.0 font processing because
     //   a single loop would involve too many checks (of font type) within
     //   the loop and slow down processing  
    else {                                 // table in 3.0 format 

        for (row = 0; row < height; row++){

             phase = 0;
             // view table as a 3.0 style table 
             pjGlyphData = work.lpFontBody;

             for (i = 0; i < (UINT)(lpFnt->CharCount); i++) {

				vGlyphInfo30FromBuffer (pjGlyphData, &gi3GlyphTable30);

                width = gi3GlyphTable30.GIwidth;

                  // size of each table element = 6bytes 
                work.lp2 = work.lpFontBody + (gi3GlyphTable30.GIoffset -
						lSizeOfOldFontHeader30 +1) + row;

				pjGlyphData += lSizeOfOldGlyph30;

                 // offset ends up as the sum of the widths 
                if (row == 0)              // Once is enough 
                    lpFnt->offsets[i + lpFnt->Header.FirstChar + 1] = offset += width;

                 // create a single scan of bitmap for character 
                phase = ConvertToBitmapFormat (&work,width, phase, height);
             }
             if ((work.lp1 - work.lpWork) & 1)
                 *work.lp1++ = 0;             // Round lp1 up to Word Boundary 
#ifdef DWORDROUND
             if ((work.lp1 - work.lpWork) & 2) {
                 *work.lp1++ = 0;             // Round lp1 up to DWord Boundary 
                 *work.lp1++ = 0;             // Round lp1 up to DWord Boundary 
			 }
#endif
        }
    }
    lpFnt->Header.WidthBytes = (WORD) (((offset + 15) >> 4) << 1);
       // fixup NEWFON error 

    GlobalFree(work.lpFontBody);
    work.lpFontBody = work.lpWork;          // So that below we free the other buffer 
    // Create a WINDOWS bitmap to move the font definition bits into 

    hDC = GetDC (equalHWND);                         // DC to be compatible with 
    lpFnt->Bitmap = CreateBitmap(
                    (INT)lpFnt->Header.WidthBytes << 3, // Width of font in pixels 
                    (INT)lpFnt->Header.PixHeight,
                    1, 1, (LPBYTE)NULL);
    lpFnt->BmDC = CreateCompatibleDC(hDC);           // Create a DC 
    SelectObject(lpFnt->BmDC, lpFnt->Bitmap);              // Relate the two of them 
    ReleaseDC(equalHWND, hDC);                      // Done with font DC 

    // Move the bits in 
    SetBitmapBits(lpFnt->Bitmap,
          (DWORD)lpFnt->Header.WidthBytes * (DWORD)lpFnt->Header.PixHeight,(CHAR  *)work.lpWork);

    // Free up the space we loaded the file into 
    GlobalFree(work.lpFontBody);
    
    SetCursor(hOldCursor);              // Restore regular cursor 
    return "";
}

// *************************************
// * compares nBytes bytes of s1 and s2 *
// *************************************
BOOL
ByteCompare (
    CHAR  *s1,
    CHAR  *s2,
    DWORD nBytes
    )
{
    for ( ; nBytes > 0; nBytes--)
        if (*s1++ != *s2++)
            return FALSE;
    return TRUE;
}

// **********************************
// * copies nBytes bytes of s2 to s1 *
// **********************************
BOOL
ByteCopy(
    CHAR  *s1,
    CHAR  *s2,
    LONG  nBytes
    )
{
    for ( ; nBytes > 0; nBytes--)
        *s1++ = *s2++;
    return TRUE;
}

// ***************************************************************************
// * void ConvertToFileFormat(width, phase, height)
// *
// * purpose  : Takes a part of the bitmap (corresponding to a single character)
// *            and converts it to a string of bytes in the font file format
// *
// * params   : WORD width :  width of the character in pixels(bits)
// *            WORD phase :  current deviation from byte alignment (pixels)
// *            WORD height:  height of character(pixels)
// *
// * returns:   WORD phase :  new deviation from byte alignment (pixels)
// *
// * side effects : modifies pointers lp1 and lp2 (pointing to font body and work
// *                space respectively)
// *
// ***************************************************************************
DWORD PASCAL
ConvertToFileFormat(WorkBufferType *lpWrk,
     DWORD width,
     DWORD phase,
     DWORD height
     )
{
      INT w;

      for (w = width; w > 0; w -= 8){  // for each byte of font 
           if (phase == 0){             // easy case 

               BYTE b;

               b = *lpWrk->lp1++;
               if (w < 8){

                   phase = (DWORD) w;
                   b >>= 8 - w;    // Clear left side bits 
                   b <<= 8 - w;
               }
               *lpWrk->lp2 = b;
           }
           else{

               DWORD j;

               lpWrk->lp1--;          // Re-read byte prevously read 
               //j = (DWORD) ((BYTE)*lp1++ << 8) | ((BYTE)*lp1++);
               j = (DWORD)((BYTE)*lpWrk->lp1++ << 8);
               j |= (DWORD) ((BYTE)*lpWrk->lp1++);
               if (w < 8){

                   j >>= 16 - phase - (w & 7);  // shove it right 
                   j <<= 8 - (w & 7); // Left justify in low byte 
                   phase += (DWORD) w;
                   if (phase <=  8)
                       lpWrk->lp1--;         // back up pointer 
                   phase &= 7;
               }
               else
                   j >>= 8 - phase;
               *lpWrk->lp2 = (BYTE)j;
           }
           lpWrk->lp2 += height;          // move to next column 
       }
       return phase;
}

// ***************************************************************************
// * char * FontSave()
// *
// * purpose: saves the work bitmap in the required font file format (2.0 or
// *          3.0 and cleans up
// *
// * params : none
// *
// * returns: ptr to a NULL string if Save goes off OK
// *          ptr to error message string otherwise
// *
// * side effects: lots
// *
// ***************************************************************************

char* FontSave(FontType *lpFnt, char *pszFileName)
{
	WorkBufferType work;
    DWORD bytecount;       // number of bytes returned by lwrite 
    DWORD size;              // total size of font 

    WORD height, row;
    DWORD i, fontlen;
    DWORD cBody, cFont,cFontsav, cFace;
    CHAR 			*lpFont;          // ponter to font body 
    CHAR * sz;
    DWORD iMax;
    WORD widthsav;
	PBYTE			pjGlyphData;
	PBYTE			pjGlyphSave;
    GLYPHINFO_20 	gi2GlyphTable20;    // 2.0 style Glyph data struct 
    GLYPHINFO_30 	gi3GlyphTable30;    // 3.0 style Glyph data struct 
	HANDLE nFile;
	DWORD BytesWritten;

	FontHeader30 font30;


    NewAverage(lpFnt); // force ave width to be recomputed 8/17/87 BobM 

    // reset file pointer 
	nFile = CreateFile(pszFileName, GENERIC_WRITE, 0,NULL,CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,NULL);

	if (nFile == INVALID_HANDLE_VALUE) 
	{
		return vszErrorOpeningFile;
	}

    // Put up an houglass ... this may take a while 
    if (!hHourGlass)
        hHourGlass = LoadCursor (NULL, IDC_WAIT);        // Get Hourglass 
    hOldCursor = SetCursor (hHourGlass);         // Show hourglass 

    height = lpFnt->Header.PixHeight;
    cBody = (DWORD)height * (DWORD)lpFnt->Header.WidthBytes;

    // Recompute file size and update header 
    if (lpFnt->FontFormat == ID_FORMAT2)
       cHeader = (WORD)(lSizeOfOldFontHeader +1);
    else
       cHeader = (WORD)(lSizeOfOldFontHeader30 - 1);

    // if of 2.0 type, check if size will exceed 64kbytes. If yes, saving
    //   in 2.0 format will result in loss of information (because offset table
    //   of 2.0 file is composed of 16bit words). Warn user and ask if file can
    //   be saved as a 3.0 file 

    if (lpFnt->FontFormat == ID_FORMAT2)
    { 	
		if (((DWORD)cHeader + (DWORD)(lSizeOfOldGlyph20 * (lpFnt->CharCount+1)) + 
			   (DWORD)cBody) >= WORD_LIMIT)
               if (MessageBox(NULL, vszTooBigFor20, vszWarning,
                              IDYES | IDCANCEL |IDNO) == IDYES)
                    lpFnt->FontFormat = ID_FORMAT3;
	}

	if (lpFnt->FontFormat == ID_FORMAT2) {

		// allocate space for a 2.0 style offsets table 
		cTable = (WORD)(lSizeOfOldGlyph20 *
				(lpFnt->CharCount + 1));

	} else {

		// allocate space for a 3.0 style offsets table 
		cTable = (WORD)(lSizeOfOldGlyph30 *
				(lpFnt->CharCount +1));
	}

	pjGlyphData = (LPSTR)GlobalAlloc (GMEM_ZEROINIT, (DWORD)cTable);

	if (pjGlyphData == 0) {

		CloseHandle(nFile);
		return vszNotEnoughMem;
	}

	pjGlyphSave = pjGlyphData;

    size = cHeader + cTable;
    iMax = lpFnt->CharCount;

    // create offsets table of font file 
    for (i = 0; i <= iMax; i++){ //HB: 1 teveel voor extra spatie van 8 breed.

         DWORD width, charSize;

         width = lpFnt->offsets[i + lpFnt->Header.FirstChar + 1] - lpFnt->offsets[i + lpFnt->Header.FirstChar];

         if (i == iMax) {
             width = 8;      // Sentinal blank 
		 }

         if (lpFnt->FontFormat == ID_FORMAT2){

            gi2GlyphTable20.GIwidth  = (SHORT)width;
            gi2GlyphTable20.GIoffset = (SHORT)size;

			vBufferFromGlyphInfo20 (&gi2GlyphTable20, pjGlyphData);

			pjGlyphData += lSizeOfOldGlyph20;

         } else {

            gi3GlyphTable30.GIwidth  = (SHORT)width;
            gi3GlyphTable30.GIoffset = (INT)size;

			vBufferFromGlyphInfo30 (&gi3GlyphTable30, pjGlyphData);

			pjGlyphData += lSizeOfOldGlyph30;
         }

         charSize = height * ((width + 7) >> 3);  // size in bytes 

         if ((size + charSize) < size)          // Overflow? 
		 {
			 GlobalFree (pjGlyphData);
			 CloseHandle(nFile);
             return vszFileTooLarge;

         }
         size += charSize;
    }

    // Update stuff in the header 

    lpFnt->Header.Face = (DWORD)size;
    cFace = (WORD)lstrlen (lpFnt->szFaceName) + 1;   // Allow for \0 
    size += cFace;
    lpFnt->Header.Size = (DWORD)size;                                // new file size 
    lpFnt->Header.BitsOffset = (DWORD)(cHeader + cTable);
    lpFnt->Header.Device = (DWORD)NULL;                   // Device Name must be NULL 
    cFontsav = size - cHeader - cTable;
    cFont =cFontsav;

	// alloc extra byte for lp1 in case it needs it 
    if (!(work.lpFontBody = (LPSTR)GlobalAlloc (GMEM_ZEROINIT, (LONG)cBody +
			height * 4))) 
	{
		GlobalFree (pjGlyphSave); //HB: (c) HB
		CloseHandle(nFile);
        return vszNotEnoughMem;
	}

    GetBitmapBits (lpFnt->Bitmap, (DWORD)cBody, work.lpFontBody);

    // save current WidthBytes 
    widthsav = lpFnt->Header.WidthBytes;

    //  MD - reset WidthBytes from computed width 
    lpFnt->Header.WidthBytes = (WORD) (cFont - cFace)/ height;

    // Allocate a block to put bitmap into 
    if ((lpFont = work.lpWork = GlobalAlloc (GMEM_ZEROINIT,(LONG)cFont)) == 0)
	{
		GlobalFree (pjGlyphSave); //HB: (c) HB
        GlobalFree (work.lpFontBody);
		CloseHandle(nFile);
        return vszNotEnoughMem;
	}

    work.lp1 = work.lpFontBody;

    // convert bitmap to file format 
    if (lpFnt->FontFormat == ID_FORMAT2){   // offsets table in 2.0 format 
        INT nChars;
        nChars =  lpFnt->CharCount;

        for (row = 0; row < height; row++){

            DWORD phase;

            phase = 0;

            pjGlyphData = pjGlyphSave;

            for (i = 0; i < (DWORD) nChars; i++) {

                INT width;

				vGlyphInfo20FromBuffer (pjGlyphData, &gi2GlyphTable20);

                width = gi2GlyphTable20.GIwidth;

                work.lp2 = (BYTE  *)(work.lpWork + (gi2GlyphTable20.GIoffset -
						cHeader - cTable + row));

				pjGlyphData += lSizeOfOldGlyph20;

                phase = ConvertToFileFormat (&work, width, phase, height);
            }
            if ((work.lp1 - work.lpWork) & 1)
                 work.lp1++;             // Round lp1 up to Word Boundary 
#ifdef DWORDROUND
            if ((work.lp1 - work.lpWork) & 2) {
                 work.lp1++;             // Round lp1 up to DWord Boundary 
                 work.lp1++;             // Round lp1 up to DWord Boundary 
			}
#endif
            //if(((offsets[lpFnt->LastChar + 1] + 7) >> 3) & 1)
                //lp1++;          // Round lp1 up to Word Boundary 
        }
    }
     // separate loops written for the 2.0 and 3.0 font processing because
     //   a single loop would involve too many checks (of font type) within
     //   the loop and slow down processing  
    else{  // table in 3.0 format 

        INT nChars;
        nChars = lpFnt->CharCount;
        for (row = 0; row < height; row++){

            DWORD phase;

            phase = 0;
			pjGlyphData = pjGlyphSave;

            for (i = 0; i < (DWORD) nChars; i++) {

                INT width;

				vGlyphInfo30FromBuffer (pjGlyphData, &gi3GlyphTable30);

                width = gi3GlyphTable30.GIwidth;

                work.lp2 = (BYTE  *)(work.lpWork + (gi3GlyphTable30.GIoffset -
						cHeader - cTable + row));

				pjGlyphData += lSizeOfOldGlyph30;

                phase = ConvertToFileFormat (&work, width, phase, height);
            }
            if ((work.lp1 - work.lpWork) & 1)
                 work.lp1++;             // Round lp1 up to Word Boundary 
#ifdef DWORDROUND
            if ((work.lp1 - work.lpWork) & 2) {
                 work.lp1++;             // Round lp1 up to DWord Boundary 
                 work.lp1++;             // Round lp1 up to DWord Boundary 
			}
#endif
            //if(((offsets[lpFnt->LastChar + 1] + 7) >> 3) & 1)
                //lp1++;          // Round lp1 up to Word Boundary 
        }
    }

	// Restore start of data. 
	pjGlyphData = pjGlyphSave;

    work.lp2 -= height - 1;              // Back up to start of character 
    for (i = 0; i < height; i++)
         *work.lp2++ = 0;             // Fill in guaranteed blank character 

    lpFnt->Header.Version = 0x200;

    //*****  code for compaction in 3.0 fmt borrowed from CMPCTFON *****
    if (lpFnt->FontFormat== ID_FORMAT3){
        DWORD iDefBitmapSize = 0;     // size of default char 
        DWORD cch = 0;                // count of number of default char 
#if 0
        GLYPHINFO_30  *cur_char;  // current element of offset table 
        GLYPHINFO_30  *move_char;
            // element from which moving is to be done 
        LONG iDefBitmapOffset;        // offset of default char 
        INT iDefBitmapWidth;          // width of default char 
        static LONG iEndOfBitmaps;    // address of end of font body 
        INT iFirstC;                  // first char in font 
        INT iLastC;                   // last char in font 
        INT i,j;
        WORD width;
        CHAR  *rgbFont;         // pointer to font body 
        DWORD Offset;

        //GDI seems to understand compressed 2.0 external formats but not
        //compressed 3.0 formats, and this is causing it to crash loading a 3.0
        //format compressed lpFnt-> Since compression does not affect the
        //validity of the font, This code is disabled temporarily till
        //this problem is verified. Again, the font is perfectly usable in this
        //form, though a trifle bigger than desired -  LR 20/26/90


		Note that this will now not work after the conversion to NT.
		the glyph table needs to accessed as above.  t-davema 8/20/91

        iFirstC = lpFnt->FirstChar & 0x0ff;
        iLastC  = lpFnt->LastChar & 0x0ff;

        iEndOfBitmaps = (LONG)lpFnt->BitsOffset+((LONG)lpFnt->WidthBytes *
                (LONG)lpFnt->PixHeight);
        rgbFont = (CHAR  *)lpFont;   // start of font body 
        cur_char =(GLYPHINFO_30  *) lpTable; // start of offset table 

        // calculate some parameters for the default char 
        iDefBitmapOffset = cur_char[lpFnt->DefaultChar].GIoffset;
        iDefBitmapWidth  = cur_char[lpFnt->DefaultChar].GIwidth;
        iDefBitmapSize   = (DWORD)cur_char[lpFnt->DefaultChar+1].GIoffset
                                           - (DWORD)iDefBitmapOffset;

        // scan the font body via the offsets table. If a default char
        //   is recognised, move all the bytes to it's right left by the
        //   size of the default char.Make all offset table entries of default
        //   char point to one image of the char (the earliest occuring image) 
        for (i = iFirstC; i <= iLastC; ++i) {

            // important: Check for limiting conditions (in case break char
               was modified?)
            if (cur_char->GIoffset == iDefBitmapOffset){
                cur_char++;
                continue;
            }

            Offset = cur_char->GIoffset -cHeader -cTable;
            // proceed to compare images only if widths are equal 
            if ((cur_char->GIwidth  == iDefBitmapWidth) &&
                (ByteCompare (&rgbFont [Offset],
                &rgbFont[iDefBitmapOffset-cHeader-cTable],
                                (DWORD)iDefBitmapSize) == TRUE)){

                // set offset to earliest occurence of the default char 
                if (cur_char->GIoffset <  iDefBitmapOffset){
                    iDefBitmapOffset = cur_char->GIoffset;
                }
                else    {
                    if (i != iLastC){
                        // move bytes to right of default char left by the
                           size of the char 
                        ByteCopy (&rgbFont [ Offset],
                                  &rgbFont [ Offset+ iDefBitmapSize],
                                  (LONG)(iEndOfBitmaps -
                                  ((cur_char + 1)->GIoffset)));

                        // correct the offset table entries 
                        move_char = cur_char + 1;
                        for (j=i; j < iLastC; ++j){
                             move_char->GIoffset -= (LONG)iDefBitmapSize;
                             move_char++;
                        }
                    }
                    iEndOfBitmaps -= iDefBitmapSize;
                    // move End-of-font to the left 
                    cur_char->GIoffset = iDefBitmapOffset;
                    // point offset of cuurent char to default char 
                    cch++;
                }
            }
            cur_char++;
        }
#endif
        // recalculate some font attributes 
        work.lp2   -= (cch * iDefBitmapSize + height);
        for (i = 0; i < height; i++)
            *work.lp2++ = 0;             // Fill in guaranteed blank character 

        cFont -= cch * iDefBitmapSize;
        lpFnt->Header.WidthBytes = (WORD) (cFont - cFace)/ height;
        lpFnt->Header.Size = (DWORD) (cFont + cHeader + cTable);
        lpFnt->Header.Face = lpFnt->Header.Size - cFace ;
        lpFnt->Header.Version = 0x300;

        // copy info into 3.0 (new) format header and set the additional
        //   fields 
        ByteCopy ((LPSTR)&font30, (LPSTR)&lpFnt->Header, (DWORD)sizeof (lpFnt->Header));
        font30.fsFlags        = 0;
        font30.fsFlags = (lpFnt->Header.Family & 1 ? FSF_PROPORTIONAL : FSF_FIXED);
        font30.fsAspace       = 0;
        font30.fsBspace       = 0;
        font30.fsCspace       = 0;
        font30.fsColorPointer = 0L;
        for (i = 0; i < 4 ; i++)
            font30.fsReserved[i] = 0L;
    }

    // Add the FaceName, if any, to the end of the bitmap 
    lstrcpy((LPSTR)work.lp2, (LPSTR)lpFnt->szFaceName);

	//
	// * Again we need to do some tricky things to get the header output
	// * correct.  We want to run it through the conversion backwards until
	// * we get the packed structure.
	 
	{
		BYTE jOutputBuffer [sizeof (font30)];

		if (lpFnt->FontFormat == ID_FORMAT2) {

			vBufferFromFontStruct (&lpFnt->Header, (PBYTE)&jOutputBuffer);

		} else {

			vBufferFromFont30Struct (&font30, (PBYTE)&jOutputBuffer);
		}

		bytecount = WriteFile(nFile, (PBYTE)&jOutputBuffer, (DWORD)cHeader,&BytesWritten,NULL) ? BytesWritten : 0;
	}

    // Write out Header Information 
    if (bytecount == cHeader) {

            // Write out OffsetsTable 
         if (cTable == 0 || (WriteFile(nFile, pjGlyphData, (DWORD)cTable,&BytesWritten,NULL) ? BytesWritten : 0)
				== (UINT)cTable)
		 {
              // Write out Body 
              fontlen = cFont;

			  if ((!WriteFile(nFile,(char *)lpFont,fontlen,&BytesWritten,NULL)) ||
				  (BytesWritten != fontlen))
			  {
				  sz = vszErrorWritingBody;
			  }
              else
			  {
                   //fChanged = FALSE;
                   sz= "";
              }
         }
         else
             sz = vszErrorWritingOffsets;
    }
    else
         sz = vszErrorWritingHdr;

    // hack: Restore saved value of widthbytes, eliminating a variety
    //   of minor scrolling problems that follow after a File/Save 
    lpFnt->Header.WidthBytes = widthsav;

	CloseHandle(nFile);

	// Tidy up 

    GlobalFree(work.lpFontBody);
    GlobalFree(pjGlyphData);
    GlobalFree(work.lpWork);

    SetCursor(hOldCursor);              // Restore regular cursor 

    return sz;
}

//***************************************************************************
// * BOOL ResizeWidths(wChar)
// *
// * params : WORD wChar : new width of a single character
// *
// * purpose: resize work bitmap according to new character width by stretching/
// *          compressing all characters as neccesary
// *
// * returns: none
// *
// * side effects: Work bitmap changes.Some header info (regarding font dimensions
// *               altered as well
// ***************************************************************************


BOOL ResizeWidths(FontType *lpFnt, DWORD wChar, HWND equalHWND)
{
    DWORD width;
    DWORD offset, i;
	ResizeBufferType Rsz;
    // Create a new bitmap to move the font definition bits into 
    width = (lpFnt->CharCount) * wChar; // In pixels 
    width = (width + 15) >> 4 << 1;     // In even bytes 
    if (!GetNewMap(lpFnt, &Rsz, width, lpFnt->Header.PixHeight,equalHWND))
        return (FALSE);

    // Move the bits in 
    offset = 0;
    Rsz.oldMode = SetStretchBltMode(Rsz.hNewMemDC, COLORONCOLOR);
    for (i = lpFnt->Header.FirstChar; i <= lpFnt->Header.LastChar; i++)
        {

        StretchBlt(Rsz.hNewMemDC, offset, 0,
                wChar, lpFnt->Header.PixHeight,                  // New character 
                lpFnt->BmDC, lpFnt->offsets[i], 0,
                lpFnt->Header.PixWidth, lpFnt->Header.PixHeight,          // Old character 
                SRCCOPY);
        lpFnt->offsets[i] = offset;
        offset += wChar;
        }
    SetStretchBltMode(Rsz.hNewMemDC, Rsz.oldMode);
    lpFnt->offsets[lpFnt->Header.LastChar + 1] = offset;

    UseNewMap(lpFnt,&Rsz);                // Switch Pointers and release the space 

    lpFnt->Header.HorizRes = (WORD) Proport(lpFnt->Header.HorizRes, wChar, lpFnt->Header.PixWidth, 999);
    lpFnt->Header.WidthBytes = (WORD) width;            // Misc. ajustments 
    lpFnt->Header.PixWidth = lpFnt->Header.AvgWidth = (lpFnt->Header.MaxWidth = (WORD) wChar);

    return (TRUE);
}


//***************************************************************************
// * BOOL SpreadWidths(wChar)
// *
// * purpose: spread/compress work bitmap (with variable width characters)
// *          in proportion with the maximum character width
// *
// * params : WORD wChar : new width of a character
// *
// * returns: none
// *
// * side effects: Work bitmap changes.Some header info (regarding font dimensions
// *               altered as well
// ***************************************************************************

BOOL SpreadWidths(FontType* lpFnt, DWORD wChar,HWND equalHWND)
{
    DWORD offset, i;
    DWORD width, oldWidth, newWidths[257];
	ResizeBufferType Rsz;

    // Create a new bitmap to move the font definition bits into 
    width = 0;          // Compute the new width 
    for (i = (DWORD) lpFnt->Header.FirstChar; i <= (DWORD) lpFnt->Header.LastChar; i++)
        {
        oldWidth = lpFnt->offsets[i + 1] - lpFnt->offsets[i];
        // Compute new width to nearest whole number 
        newWidths[i] = Proport(oldWidth, wChar, lpFnt->Header.MaxWidth, wBoxLim - 1);
        width += newWidths[i];
        }
    width = (width + 15) >> 4 << 1;     // In even bytes 
    if (!GetNewMap(lpFnt, &Rsz, width, lpFnt->Header.PixHeight,equalHWND))
        return(FALSE);

    // Move the bits in 
    offset = 0;
    Rsz.oldMode = SetStretchBltMode(Rsz.hNewMemDC, COLORONCOLOR);
    for (i = lpFnt->Header.FirstChar; i <= lpFnt->Header.LastChar; i++)
        {
        oldWidth = lpFnt->offsets[i + 1] - lpFnt->offsets[i];
        StretchBlt(Rsz.hNewMemDC, offset, 0,
                newWidths[i], lpFnt->Header.PixHeight,           // New character 
                lpFnt->BmDC, lpFnt->offsets[i], 0,
                oldWidth, lpFnt->Header.PixHeight,               // Old character 
                SRCCOPY);
        lpFnt->offsets[i] = offset;
        offset += newWidths[i];
        }
    SetStretchBltMode(Rsz.hNewMemDC, Rsz.oldMode);
    lpFnt->offsets[lpFnt->Header.LastChar + 1] = offset;

    UseNewMap(lpFnt,&Rsz);                // Switch Pointers and release the space 
    lpFnt->Header.HorizRes = (WORD) Proport(lpFnt->Header.HorizRes, wChar, lpFnt->Header.MaxWidth, 999);
    lpFnt->Header.WidthBytes = (WORD) width;            // Misc. ajustments 
    lpFnt->Header.MaxWidth = (WORD) wChar;
    NewAverage(lpFnt);                       // Compute new average width 

    return (TRUE);
}


//***************************************************************************
// * NewAverage()
// *
// * purpose: recalculate average width of a character in font
// *
// * params : none
// *
// * returns: none
// *
// * side effects: alters the average width parameter in font header
// *
// ***************************************************************************
void NewAverage(FontType *lpFnt)
{
#ifdef FOO
    WORD i, totalwidth;
    // 12/23/85 -- use weighted avg of lower case letters (mikecr) 

    // width of the space 
    totalwidth = (offsets[' ' + 1] - offsets[' ']) * widthweights[ALPHA_CNT];

    for (i = 0; i < ALPHA_CNT; i++)
            totalwidth += (offsets['a' + i + 1] - offsets['a' + i]) *
            widthweights[i];

    lpFnt->AvgWidth = totalwidth / TOTAL_WEIGHTS;

    // round up if necessary 
    if (totalwidth % TOTAL_WEIGHTS >= (TOTAL_WEIGHTS >> 1))
            lpFnt->AvgWidth++;

#endif

    // lets do a simple average here 
    lpFnt->Header.AvgWidth = (WORD) (((lpFnt->offsets[lpFnt->Header.LastChar+1] -
            lpFnt->offsets[lpFnt->Header.FirstChar]) + (lpFnt->CharCount-1)/2) /
            (lpFnt->CharCount));

    if (lpFnt->Header.AvgWidth == 0) {
       lpFnt->Header.AvgWidth++;
    }
}


//***************************************************************************
// * BOOL ResizeBody(width, height)
// *
// * purpose: adjust work bitmap according to new specified dimensions
// *
// * params : WORD width  : new width of font in pixels
// *          WORD height : new height of font in pixels
// *
// * returns: none
// *
// * side effects: Work bitmap changes.Some header info (regarding font dimensions
// *               altered as well
// ***************************************************************************

BOOL ResizeBody(FontType* lpFnt, DWORD width, DWORD height,HWND equalHWND)
{
	ResizeBufferType Rsz;
    // Create a new bitmap to move the font definition bits into 
    if (!GetNewMap(lpFnt, &Rsz, width, height,equalHWND))
        return(FALSE);

    // Move the bits in 
    Rsz.oldMode = SetStretchBltMode(Rsz.hNewMemDC, COLORONCOLOR);
    StretchBlt(Rsz.hNewMemDC, 0, 0,
            width << 3, height,         // New Char. 
            lpFnt->BmDC, 0, 0,
            width << 3, lpFnt->Header.PixHeight, // Old Char. 
            SRCCOPY);
    SetStretchBltMode(Rsz.hNewMemDC, Rsz.oldMode);

    UseNewMap(lpFnt,&Rsz);                // Switch Pointers and release the space 

    lpFnt->Header.ExtLeading = (WORD) Proport(lpFnt->Header.ExtLeading, height, lpFnt->Header.PixHeight, 999);
    lpFnt->Header.IntLeading = (WORD) Proport(lpFnt->Header.IntLeading, height, lpFnt->Header.PixHeight, 999);
    lpFnt->Header.Ascent = (WORD) Proport(lpFnt->Header.Ascent, height, lpFnt->Header.PixHeight, 32);
    lpFnt->Header.VertRes = (WORD) Proport(lpFnt->Header.VertRes, height, lpFnt->Header.PixHeight, 999);
    lpFnt->Header.Points = (WORD) Proport(lpFnt->Header.Points, height, lpFnt->Header.PixHeight, 999);
    lpFnt->Header.PixHeight = (WORD) height;            // Fix misc. header values 
    lpFnt->Header.WidthBytes = (WORD) width;

    return (TRUE);
}


//***************************************************************************
// * BOOL NewFirstChar(first)
// *
// * purpose: redefines first character in font and resizes work bitmap
// *          accordingly
// *
// * params : WORD first : new first character to be defined
// *
// * returns: none
// *
// * side effects: Work bitmap changes.Some header info (regarding font dimensions
// *               altered as well
// ***************************************************************************

BOOL NewFirstChar(FontType* lpFnt, char first, HWND equalHWND)
{
    DWORD  width, wDefault;
    DWORD offset, i;
    INT dw;
	ResizeBufferType Rsz;

    if (first > lpFnt->Header.FirstChar)         // Smaller? 
        {
        ShrinkFont(lpFnt, first, lpFnt->Header.LastChar, equalHWND);
        lpFnt->Header.FirstChar = (BYTE) first;
        //return(FALSE);
        return(TRUE);
        }

    // If not smaller we must pad with the default character 
    wDefault = lpFnt->offsets[lpFnt->Header.DefaultChar + 1] - lpFnt->offsets[lpFnt->Header.DefaultChar];
    dw = wDefault * (lpFnt->Header.FirstChar - first);           // Extra width 
    width = lpFnt->offsets[lpFnt->Header.LastChar + 1] + dw;    // New width (pixels) 
    width = (width + 15) >> 4 << 1;             // Width - even bytes 
    if (!GetNewMap(lpFnt, &Rsz, width, lpFnt->Header.PixHeight,equalHWND))
        return(FALSE);           // New work area 

    // Move it in in two parts 
    // First move in default characters 
    offset = 0;
    for (i = first; i < lpFnt->Header.FirstChar; i++)
        {
        BitBlt(Rsz.hNewMemDC, offset, 0,
                wDefault, lpFnt->Header.PixHeight,
                lpFnt->BmDC, lpFnt->offsets[lpFnt->Header.DefaultChar], 0,
                SRCCOPY);
        lpFnt->offsets[i] = offset;
        offset += wDefault;
        }
    // Now move in the rest 
    BitBlt(Rsz.hNewMemDC, offset, 0,
            lpFnt->offsets[lpFnt->Header.LastChar + 1], lpFnt->Header.PixHeight,
            lpFnt->BmDC, 0, 0,
            SRCCOPY);

    UseNewMap(lpFnt,&Rsz);                // Switch Pointers and release the space 

    // Now fix up offsets table 
    for (i = lpFnt->Header.FirstChar; i <= (DWORD)(lpFnt->Header.LastChar + 1); i++)
        lpFnt->offsets[i] = lpFnt->offsets[i] + dw;           // Shift the rest right 
    lpFnt->Header.WidthBytes = (WORD) width;
    lpFnt->Header.FirstChar = (BYTE) first;

    return (TRUE);
}


//***************************************************************************
// * ShrinkFont(first, last)
// *
// * purpose:  redefine the first and last charcter in the font and shrink
// *           work bitmap accordingly
// *
// * params :  WORD first : new first character to be defined
// *           WORD last  : new last character  "  "   "
// *
// * returns:  none
// *
// * side effects: Work bitmap changes.Some header info (regarding font dimensions
// *               altered as well
// ***************************************************************************

void ShrinkFont(FontType *lpFnt, char first, char last, HWND equalHWND)
{
    DWORD width, widthPixels;
    char i;
    INT dw;
	ResizeBufferType Rsz;

    dw = lpFnt->offsets[first] - lpFnt->offsets[lpFnt->Header.FirstChar];      // left shift if any 
    widthPixels = lpFnt->offsets[last + 1] - lpFnt->offsets[first];   // Width in pixels 
    width = (widthPixels + 15) >> 4 << 1;               // Width - even bytes 
    if (!GetNewMap(lpFnt, &Rsz, width, lpFnt->Header.PixHeight, equalHWND))
        return;           // New work area.

    // Now move the font into the reduced space 

    BitBlt(Rsz.hNewMemDC, 0, 0,
            widthPixels, lpFnt->Header.PixHeight,
            lpFnt->BmDC, lpFnt->offsets[first], 0,
            SRCCOPY);

    UseNewMap(lpFnt, &Rsz);                // Switch Pointers and release the space 

    if (dw)                             // Ajust offsets 
        {
        for (i = first; i <= last + 1; i++)
            lpFnt->offsets[i] -= dw;
        }

    lpFnt->Header.WidthBytes = (WORD) width;

}


//***************************************************************************
// * BOOL NewLastChar(last)
// *
// * purpose:  redefines the last character in the font
// *
// * params :  WORD last : number of character to be made the last character
// *
// * returns:  none
// *
// * side effects: Work bitmap changes.Some header info (regarding font dimensions
// *               altered as well
// *
// ***************************************************************************

BOOL NewLastChar(FontType* lpFnt, char last, HWND equalHWND)
{
    DWORD  width, wDefault;
    DWORD offset;
	char i;
    INT dw;
	ResizeBufferType Rsz;

    if (last < lpFnt->Header.LastChar)           // Smaller? 
        {
        ShrinkFont(lpFnt, lpFnt->Header.FirstChar, last, equalHWND);
        lpFnt->Header.LastChar = (BYTE) last;
        return(FALSE);
        }

    // If not smaller we must pad with the default character 
    wDefault = lpFnt->offsets[lpFnt->Header.DefaultChar + 1] - lpFnt->offsets[lpFnt->Header.DefaultChar];
    dw = wDefault * (last - lpFnt->Header.LastChar);             // Extra width 
    offset = lpFnt->offsets[lpFnt->Header.LastChar + 1];        // Current end 
    width = offset + dw;                        // New width (pixels) 
    width = (width + 15) >> 4 << 1;             // Width - even bytes 
    if (!GetNewMap(lpFnt,&Rsz, width, lpFnt->Header.PixHeight, equalHWND))
        return(FALSE);           // New work area 

    // Move it in in two parts 
    // First move in the existing font 
    BitBlt(Rsz.hNewMemDC, 0, 0,
            offset, lpFnt->Header.PixHeight,
            lpFnt->BmDC, 0, 0,
            SRCCOPY);
    // Then move in default characters 
    for (i = lpFnt->Header.LastChar + 1; i <= last;)
        {
        BitBlt(Rsz.hNewMemDC, offset, 0,
                wDefault, lpFnt->Header.PixHeight,
                lpFnt->BmDC, lpFnt->offsets[lpFnt->Header.DefaultChar], 0,
                SRCCOPY);
        offset += wDefault;
        lpFnt->offsets[++i] = offset;
        }

    UseNewMap(lpFnt, &Rsz);                // Switch Pointers and release the space 

    lpFnt->Header.WidthBytes = (WORD) width;
    lpFnt->Header.LastChar = (BYTE) last;

    return (TRUE);
}


//***************************************************************************
// * BOOL CharWidth(iChar, wBox)
// *
// * purpose: resizes selected char according to new dimensions. (only for
// *          variable pitch)
// *
// * params : BYTE iChar : character to resize
// *          WORD wBox  : new width of char in pixels
// *
// * returns: none
// *
// * side effects: work bitmap pixel values and header info(regarding font
// *               dimensions) altered
// *
// ***************************************************************************

BOOL CharWidth(FontType* lpFnt, char iChar, DWORD wBox, HWND equalHWND)
{
    DWORD  width, nChars;
    DWORD  w1, w2, i;
    INT dw;
	ResizeBufferType Rsz;

    nChars = lpFnt->CharCount;        // Character count 
    dw = wBox - (lpFnt->offsets[iChar + 1] - lpFnt->offsets[iChar]);  // Width change 
    width = lpFnt->offsets[lpFnt->Header.LastChar + 1] + dw;            // New width (pixels) 
    width = (width + 15) >> 4 << 1;                     // Width - even bytes 
    if (!GetNewMap(lpFnt, &Rsz, width, lpFnt->Header.PixHeight, equalHWND))
        return(FALSE);                   // New work area 

    // Move it in in two parts 
    // First move up to and including iChar 
    w1 = lpFnt->offsets[iChar + 1];            // Width (in pixels) to move 
    BitBlt(Rsz.hNewMemDC, 0, 0,
            w1 + dw, lpFnt->Header.PixHeight,
            lpFnt->BmDC, 0, 0,
            SRCCOPY);
    // Now move in the rest 
    if (iChar < (BYTE) lpFnt->Header.LastChar)        // Part to right of elision 
        {
        w2 = lpFnt->offsets[lpFnt->Header.LastChar + 1] - lpFnt->offsets[iChar + 1];
        BitBlt(Rsz.hNewMemDC, lpFnt->offsets[iChar] + wBox, 0,
                w2, lpFnt->Header.PixHeight,
                lpFnt->BmDC, lpFnt->offsets[iChar + 1], 0,
                SRCCOPY);
        }

    UseNewMap(lpFnt, &Rsz);                // Switch Pointers and release the space 

    // Now fix up offsets table 
    for (i = iChar + 1;                         // Where changes start 
        i <= (DWORD)(lpFnt->Header.LastChar + 1); i++)            // Ajust offsets 
        lpFnt->offsets[i] = lpFnt->offsets[i] + dw;           //  .. by adding dw 
    lpFnt->Header.WidthBytes = (WORD) width;
    NewAverage(lpFnt);

    return (TRUE);
}

//***************************************************************************
// * GetNewMap(width, height)
// *
// * purpose: create new bitmap of the given width and height.
// *
// * params : WORD width  : width of bitmap in pixels
// *          WORD height : height of bitmap in pixels
// *
// * returns: TRUE if successful, FALSE otherwise
// *
// * side effects: Handle and DC values of new DC assigned
// *
// ***************************************************************************

BOOL GetNewMap(FontType* lpFnt, ResizeBufferType *lpRsz, DWORD width, DWORD height,HWND equalHWND)
{
    HDC hDC;

    if (height==0) // Check if something stupid is happening 
        height=lpFnt->Header.PixHeight; // Fix it 

    // Put up an houglass ... this may take a while 
    if (!hHourGlass)
        hHourGlass = LoadCursor(NULL, IDC_WAIT);        // Get Hourglass 
    hOldCursor = SetCursor(hHourGlass);         // Show hourglass 

    // Create a new bitmap to move the font definition bits into 
    hDC = GetDC(equalHWND);                         // DC to be compatible with 
    lpRsz->hNewBitmap = CreateBitmap(
                    width << 3,                 // Width of font in pixels 
                    height,
                    1, 1, (LPBYTE)NULL);
    if (!lpRsz->hNewBitmap)
	{
		MessageBox(NULL,vszNotEnoughMem,"Error",MB_APPLMODAL | MB_ICONERROR | MB_OK);
        ReleaseDC(equalHWND, hDC);    // bug# 2380 
        return FALSE;
        }

    lpRsz->hNewMemDC = CreateCompatibleDC(hDC);						// Create a DC 
    SelectObject(lpRsz->hNewMemDC, lpRsz->hNewBitmap);				// Relate them 
    ReleaseDC(equalHWND, hDC);										// Done with font DC 
    PatBlt(lpRsz->hNewMemDC, 0, 0, width << 3, height, BLACKNESS);  // Clear it 
    return TRUE;
}


//***************************************************************************
// * UseNewMap()
// *
// * params : none
// *
// * purpose: discard old bitmap and replace it with new one
// *
// * returns: none
// *
// * side effects: Handle to old bitmap and handle to old bitmap DC replaced
// *               by those of new bitmap respectively
// *
// ***************************************************************************

void UseNewMap(FontType* lpFnt, ResizeBufferType *lpRsz)
{
    DeleteDC(lpFnt->BmDC);
    DeleteObject(lpFnt->Bitmap);              // Release old space 
    lpFnt->Bitmap = lpRsz->hNewBitmap;
    lpFnt->BmDC = lpRsz->hNewMemDC;              // Release old space 
    //SetCursor(hOldCursor);              // Restore regular cursor 
}



void DeleteGlobalBitmap(FontType* lpFnt)
{
    if (lpFnt->BmDC)
        DeleteDC(lpFnt->BmDC);
    if (lpFnt->Bitmap)
        DeleteObject(lpFnt->Bitmap);
	lpFnt->FontFormat = ID_NOFONT;
}


DWORD
Proport(
    DWORD value,
    DWORD top,
    DWORD bottom,
    DWORD limit
    )
    // Reproportion a value by the ratio top/bottom to the nearest integer
    // * and make sure we are still in range 
{
    return min(limit, (DWORD)((1 + 2 * value * top) / (2 * bottom)));
}

/****************************************************************************
 * char * VerifyHeaderContents()
 *
 * purpose: Checks if the Header information of the file just read makes
 *          sense. If not, returns an error message string to FontLoad
 *
 * params : none
 *
 * returns :char *szError : ptr to error string if error occurs
 *                          NULL otherwise
 *
 * side effects: none
 *
 ***************************************************************************/


char* PASCAL VerifyHeaderContents(FontHeaderType* lpFnt)
{
     if (lpFnt->Points > 999)
        return vszNomPtSizeNotOk;
     if (lpFnt->VertRes > 999)
        return vszNomVertResNotOk;
     if (lpFnt->HorizRes > 999)
        return vszNomHorResNotOk;
     if (lpFnt->Ascent > lpFnt->PixHeight)
        return vszAscentTooBig;
     if (lpFnt->Ascent > 999)
        return vszAscentNotOk;
     if (lpFnt->ExtLeading > 999)
        return vszExtLeadNotOk;
     if (lpFnt->IntLeading > lpFnt->Ascent)
        return vszIntLeadTooBig;
     if (lpFnt->IntLeading > 999)
        return vszIntLeadNotOk;
     if (lpFnt->LastChar > 255)
        return vszCharSetOutOfBounds;
     if (lpFnt->DefaultChar > lpFnt->FirstChar + (lpFnt->LastChar - lpFnt->FirstChar))
#if 0
          lpFnt->DefaultChar = 0;
#else
        return vszDefCharOutsideFont;
#endif
     if (lpFnt->DefaultChar > 255)
        return vszDefCharNotOk;
     if (lpFnt->BreakChar > (lpFnt->LastChar - lpFnt->FirstChar))
        return vszBreakCharOutsideFont;
     if (lpFnt->BreakChar > 255)
        return vszBreakCharNotOk;
     if (lpFnt->PixHeight > 64) 
        return vszHeightOutOfBounds;
     if (lpFnt->MaxWidth > 64)
        return vszMaxWidthOutOfBounds;
     if (lpFnt->AvgWidth > 64)
        return vszAvgWidthOutOfBounds;
#define SEGMENT_SIZE 65536
     if (lpFnt->Version == ID_FORMAT2)
        if (lpFnt->BitsOffset > (DWORD)SEGMENT_SIZE)
            return vszBitsOffsetNotOk;
#undef SEGMENT_SIZE
     return NULL;
}


#if 0
/****************************************************************************
 * BOOL PASCAL Format20FileTooBig(iHeight, iWidth)
 *
 * purpose: Checks if the 2.0 font file size is within the 64k
 *          limit imposed by the 2.0 format offset table.( if file over
 *          size limit is saved  in 2.0 format, loss of information will
 *          result)
 *
 * params:  WORD iHeight : current height of font in pixels
 *          WORD iWidth  : current width of font in pixels
 *
 * returns: TRUE  : file too large
 *          FALSE : ok to save
 *
 * side effects: none
 *
 ***************************************************************************/
BOOL PASCAL 
Format20FileTooBig(
	WORD iHeight,
	WORD iWidth
	)
{
    if ((DWORD)lSizeOfOldFontHeader + (DWORD)cTable
         + (lpFnt->LastChar - lpFnt->FirstChar +1)
         * (((DWORD)iHeight * (DWORD)iWidth ) >> 3) >= WORD_LIMIT)
             return TRUE;
    return FALSE;
}
#endif


